-------------------------------------------------------------------------------
-- Title      : Connection ID Table
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : ConnIDTable.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/10
-------------------------------------------------------------------------------
-- Description: The Connection ID Table is a simple memory of size 16 that
-- stores the injection IDs or the virtual channla numbers for incoming and
-- outgoing ports. Each entry of the memory is 32 bits. 
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;

entity ConnIDTable is
  port (
    clk           : in  std_logic;
    reset         : in  std_logic;
    read_address  : in  std_logic_vector(31 downto 0);
    read_data     : out std_logic_vector(31 downto 0);
    write_address : in  std_logic_vector(31 downto 0);
    write_data    : in  std_logic_vector(31 downto 0);
    MemWrite      : in  std_logic;
    MemRead       : in  std_logic
    );
end ConnIDTable;

architecture behavioral of ConnIDTable is

  type mem_array is array(0 to 15) of std_logic_vector (31 downto 0);



begin

  mem_process         : process(clk, reset)
    variable data_mem : mem_array := (
      X"00000000",                      -- initialize data memory
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000",
      X"00000000"
      );
    variable rd_addr  : integer;
    variable wr_addr  : integer;

    function to_integer(X : std_logic_vector) return integer is
      variable result     : integer;
    begin
      result                       := 0;
      for i in X'range loop
        result                     := result * 2;
        case X(i) is
          when '0' | 'L' => null;
          when '1' | 'H' => result := result + 1;
          when others    => null;
        end case;
      end loop;
      return result;
    end to_integer;

  begin
    if reset = '1' then
      read_data <= (others     => '0');
      for i in 0 to 15 loop
        data_mem(i) := (others => '0');
      end loop;
    elsif rising_edge(clk) then
      rd_addr             := to_integer(read_address(3 downto 0));
      wr_addr             := to_integer(write_address(3 downto 0));
      if (MemWrite = '1') then
        data_mem(wr_addr) := write_data;
      end if;
      if (MemRead = '1') then
        read_data <= data_mem(rd_addr) after 10 ns;
      end if;
    end if;
  end process mem_process;
end behavioral;

library IEEE;
use IEEE.std_logic_1164.all;

package ConnIDTable_pkg is
  component ConnIDTable
    port (
      clk           : in  std_logic;
      reset         : in  std_logic;
      read_address  : in  std_logic_vector(31 downto 0);
      read_data     : out std_logic_vector(31 downto 0);
      write_address : in  std_logic_vector(31 downto 0);
      write_data    : in  std_logic_vector(31 downto 0);
      MemWrite      : in  std_logic;
      MemRead       : in  std_logic);
  end component;
end ConnIDTable_pkg;
