-------------------------------------------------------------------------------
-- Title      : Decapsulation Unit
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : Decap.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/10
-------------------------------------------------------------------------------
-- Description: The Decapsulation Unit is responsible for choosing packets to
-- depack from among the 8 incoming virtual channels. It assembles the packet,
-- deciphers its type, extracts the contents of the packet, and present the
-- contents to either the ReqE2E unit if the packet is a response packet, or to
-- the RespE2E unit if the packet is a request packet. 
-------------------------------------------------------------------------------



library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;
use ieee.numeric_std.all;

entity Decap is
  port (
    clk              : in  std_logic;
    reset            : in  std_logic;
    SBA              : in  std_logic_vector(SBA_length-1 downto 0);
    -- Network Interface signals
    -- incoming signal indicating whether the indexed VC is empty or not.
    -- Active High.
    empty            : in  std_logic_vector(num_vc-1 downto 0);
    -- outgoing control signal to each of the virtual channels. It indicates
    -- whether a flit is taken from the indexed virtual channel or not. 
    get              : out std_logic_vector(num_vc-1 downto 0);
    -- mux control signal
    inChSel          : out std_logic_vector(chSelect-1 downto 0);
    -- incoming flit data.
    inFlit           : in  std_logic_vector(flit_size-1 downto 0);
    -- BE packet header signal 
    routepath        : out std_logic_vector(routepath_length-1 downto 0);
    MAddr            : out std_logic_vector(MAddr_length-1 downto 0);
    MCmd             : out std_logic_vector(MCmd_length-1 downto 0);
    packet_decode    : out Packet_Type;
    -- Outgoing Request Signals
    MFlag            : out std_logic_vector(MFlag_length-1 downto 0);
    MData            : out std_logic_vector(MData_length-1 downto 0);
    MDataInfo        : out std_logic_vector(MDataInfo_length-1 downto 0);
    Request_arrived  : out std_logic;
    -- Outgoing Response Signals
    SResp            : out std_logic_vector(SResp_length-1 downto 0);
    SData            : out std_logic_vector(SData_length-1 downto 0);
    SDataInfo        : out std_logic_vector(SDataInfo_length-1 downto 0);
    Response_arrived : out std_logic);
end entity Decap;


architecture behavioral of Decap is


  ------------------------------------------------------------------------------
  -- These mask constants are used to differentiate the positions of the 
  ------------------------------------------------------------------------------
  constant GT_mask   : std_logic_vector(num_vc-1 downto 0) := "11000000";
  constant GP_mask   : std_logic_vector(num_vc-1 downto 0) := "00110000";
  constant BE_H_mask : std_logic_vector(num_vc-1 downto 0) := "00001100";
  constant BE_L_mask : std_logic_vector(num_vc-1 downto 0) := "00000011";

  -- Signal Delaration
  type inflitRegType is array(flit_num) of std_logic_vector(flit_size -2 downto 0);
  signal in_flit_array                   : inflitRegType;
  type maskarray_type is array(3 downto 0) of std_logic_vector(num_vc-1 downto 0);
  signal mask_array                      : maskarray_type;
  signal regs                            : RegType;  
  signal get_check                       : std_logic;
  signal finalGet                        : std_logic_vector(num_vc-1 downto 0) := (others => '0');
  signal intermediateGet                 : std_logic_vector(num_vc-1 downto 0);
  signal intermediatePri                 : priority;
  signal intermediateSel                 : std_logic_vector(chSelect-1 downto 0);
  signal packet_received                 : std_logic                           := '0';
  signal inChPri                         : Priority;
  type get_array is array(7 downto 0) of std_logic_vector(7 downto 0);
  type sel_array is array(7 downto 0) of integer range 0 to 7;
  signal get_queue                       : get_array;
  signal sel_queue                       : sel_array;
  signal take_tag                        : std_logic_vector(7 downto 0);
  signal insert                          : integer range 0 to num_vc           := 0;
  signal take                            : integer range 0 to num_vc           := 0;
  signal Request_Valid, Response_Valid   : std_logic;
  signal take_counter, take_counter_next : integer range 0 to num_vc           := 0;
  type state_type is (init, keep, change);
  signal cs, ns                          : state_type;


begin  -- behavioral
-------------------------------------------------------------------------------
-- Name: Incoming Virtual Channel Priority Table
-- Purpose: Allow the priority scheduler to know the priority of each of the
-- incoming virtual channels. This table is currently set statically. 
-------------------------------------------------------------------------------
  regs(0) <= BE_L;
  regs(1) <= BE_L;
  regs(2) <= BE_H;
  regs(3) <= BE_H;
  regs(4) <= GP;
  regs(5) <= GP;
  regs(6) <= GT;
  regs(7) <= GT;

-------------------------------------------------------------------------------
-- Name: Incoming Virtual Channel Priority Masks
-- Purpose: Priority masks are used by SelectChannel process when it needs to
-- differentiate channels of the same priority level apart from others. 
-------------------------------------------------------------------------------
  mask_array(3) <= GT_mask;
  mask_array(2) <= GP_mask;
  mask_array(1) <= BE_H_mask;
  mask_array(0) <= BE_L_mask;

-------------------------------------------------------------------------------
-- Name: Select Channel
-- Purpose: Based on the priorities of the incoming virtual channels assigned
-- in the Priority Table and the empty bits for each channel, the Select Channel
-- process decides from which incoming virtual channel to select the next incoming packet.
-------------------------------------------------------------------------------
  SelectChannel               : process (mask_array, regs, empty)
    variable channel_selected : integer;
    variable pri              : integer range 0 to 3                := 0;
    variable ch_num, k        : integer range 0 to num_vc-1         := 0;
    variable contain_flit     : std_logic_vector(7 downto 0);
    variable temp_untagged    : std_logic_vector(num_vc-1 downto 0) := (others => '0');
    -- tag indicating which virtual channel has previously been selected for each
    -- priority type.
    variable tag              : std_logic_vector(num_vc-1 downto 0) := (others => '0');
    variable zero             : std_logic_vector(num_vc-1 downto 0) := (others => '0');
  begin  -- process
    -- The 'pri' for loop traverse through each priority level from highest
    -- priority to lowest priority.
    channel_selected                                                := 0;
    for pri in 3 downto 0 loop
      contain_flit                                                  := (others => '0');
      if channel_selected = 0 then
        -- box 1
        for ch_num in 0 to num_vc-1 loop
          if regs(ch_num) = pri and empty(ch_num) = '0' then
            contain_flit(ch_num)                                    := '1';
            channel_selected                                        := channel_selected + 1;
          end if;
        end loop;
        next when channel_selected = 0;
        -- box 4
        if temp_untagged /= zero then
          for ch_num in 0 to num_vc-1 loop
            if temp_untagged(ch_num) = '1' then
              intermediateGet         <= (others                               => '0');
              intermediateGet(ch_num) <= '1';
              temp_untagged(ch_num)                                 := '0';
              tag                                                   := tag and not mask_array(pri);
              tag(ch_num)                                           := '1';
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        -- box 2
        elsif channel_selected = 1 then
          intermediateGet             <= contain_flit;
          tag                                                       := (tag and not mask_array(pri)) or contain_flit;
          for ch_num in 0 to num_vc-1 loop
            if contain_flit(ch_num) = '1' then
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        -- box 3
        else
          temp_untagged                                             := contain_flit and not tag;
          for ch_num in 0 to num_vc-1 loop
            if temp_untagged(ch_num) = '1' then
              for k in 0 to 7 loop
                if ch_num = sel_queue(k) and take_tag(k) = '0' then
                  if (k < insert) or (k > insert and insert < take_counter) then
                    temp_untagged(ch_num)                           := '0';
                  end if;
                end if;
              end loop;  -- k
            end if;
          end loop;
          for ch_num in 0 to num_vc-1 loop
            if temp_untagged(ch_num) = '1' then
              intermediateGet         <= (others => '0');
              intermediateGet(ch_num) <= '1';
              temp_untagged(ch_num)   := '0';
              tag                     := tag and not mask_array(pri);
              tag(ch_num)             := '1';
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        end if;
      end if;
    end loop;
    if channel_selected = 0 then
      intermediateGet                 <= (others => '0');
      intermediateSel                 <= (others => '0');
    end if;
  end process;

-------------------------------------------------------------------------------
-- Name: Priority Scheduler Finite State Machine
-- Purpose: Controls the selecting of virtual channels such that it makes sure
-- a channel is selected for consecutive clock cycles until all the flits from
-- a packet has been received. Once the full packet is received, the state
-- machine will select the next channel lined up in the select and get queue. 
-------------------------------------------------------------------------------
  FSM_register : process (clk, reset)
  begin  -- process FSM
    if reset = '1' then                 -- asynchronous reset (active high)
      cs           <= init;
      take_counter <= 0;
    elsif clk'event and clk = '1' then  -- rising clock edge
      cs           <= ns;
      take_counter <= take_counter_next;
    end if;
  end process FSM_register;

  FSM_combinatorial         : process (cs, intermediateGet, intermediateSel, intermediatePri, packet_received, take_counter, get_queue, sel_queue, regs)
    variable   restart_flag : boolean                                           := false;
    variable   ones         : std_logic_vector(7 downto 0)                      := (others => '1');
    ---------------------------------------------------------------------------
    -- Function: to_int
    -- Purpose : converts an unsigned std_logic_vector to an integer
    ---------------------------------------------------------------------------
    function to_int(X       : std_logic_vector) return integer is
      variable result       : integer;
    begin
      result                                                                    := 0;
      for i in X'range loop
        result                                                                  := result * 2;
        case X(i) is
          when '0' | 'L'                                              => null;
          when '1' | 'H'                                              => result := result + 1;
          when others                                                 => null;
        end case;
      end loop;
      return result;
    end function to_int;
  begin  -- process FSM_next
    ns                      <= cs;
    case cs is
      when init                                                       =>
        if to_int(intermediateGet) = 0 then
          finalGet          <= (others                                => '0');
          inChSel           <= (others                                => '0');
          inChPri           <= BE_L;
          ns                <= init;
        else
          finalGet          <= intermediateGet;
          inChSel           <= intermediateSel;
          inChPri           <= intermediatePri;
          ns                <= keep;
        end if;
      when keep                                                       =>
        if packet_received = '0' then
          ns                <= keep;
        else
          ns                <= change;
        end if;
        if empty = ones then
          restart_flag                                                          := true;
        end if;
        if restart_flag = true and empty /= ones then
          ns                <= change;
          restart_flag                                                          := false;
        end if;
      when change                                                     =>
        finalGet            <= get_queue(take_counter);
        inChSel             <= std_logic_vector(to_unsigned(sel_queue(take_counter), chSelect));
        inChPri             <= regs(sel_queue(take_counter));
        take_counter_next   <= take_counter + 1;
        if take_counter = 7 then
          take_counter_next <= 0;
        end if;
        ns                  <= keep;
      when others                                                     => null;
    end case;
  end process FSM_combinatorial;

  get <= finalGet;

  
 -------------------------------------------------------------------------------
-- Name: Get FIFO and Sel FIFO
-- Purpose: To store and keep track of all the selected channels in order as
-- the decision were make by Select Channel process such
-- that the decisions will not be lost while the finite state machine is in the
-- "keep" state and not ready to pick a new channel yet. Get queue stores all
-- the "get" values for all 8 virtual channels. Sel queue stores all the "inChSel"
-- values for each get signal. 
-------------------------------------------------------------------------------
  FIFO                      : process (clk, reset)
    variable insert_counter : integer range 0 to num_vc := 0;
  begin  -- process
    if reset = '1' then
      take_tag                          <= (others => '1');
      for i in 0 to 7 loop
        get_queue(i)                    <= (others => '0');
      end loop;  -- i
    elsif rising_edge(clk) then
      if insert_counter = 0 and intermediateGet /= finalGet then
        if take_tag(insert_counter) = '1' then   -- if taken 
          get_queue(insert_counter)     <= intermediateGet;
          for i in 0 to num_vc -1 loop
            if intermediateGet(i) = '1' then
              sel_queue(insert_counter) <= i;
            end if;
          end loop;  -- i 
          take_tag(insert_counter)      <= '0';  -- set not taken
          insert_counter                                := insert_counter + 1;
        end if;
      elsif insert_counter > 0 and get_queue(insert_counter-1) /= intermediateGet then
        if take_tag(insert_counter) = '1' then
          get_queue(insert_counter)     <= intermediateGet;
          for i in 0 to num_vc -1 loop
            if intermediateGet(i) = '1' then
              sel_queue(insert_counter) <= i;
            end if;
          end loop;  -- i 
          take_tag(insert_counter)      <= '0';
          insert_counter                                := insert_counter + 1;
          if insert_counter = 8 then
            insert_counter                              := 0;
          end if;
        end if;
      end if;
      insert                            <= insert_counter;
      if cs = change then
        take_tag(take_counter)          <= '1';  --set taken
      end if;
    end if;
  end process;

  get_check <= finalGet(0) or finalGet(1) or finalGet(2) or finalGet(3) or finalGet(4) or finalGet(5) or finalGet(6) or finalGet(7);

 -------------------------------------------------------------------------------
-- Name: Packet Assemble
-- Purpose: To assemble all the incoming flits to a its original packet and send
-- to the finite state machine a "packet received" signal when the full packet
-- has been received. Therefore, allowing the finite state machine to select a
-- different channel if they are lined up in the Sel and Get queues. 
-------------------------------------------------------------------------------
  Packet_Assemble       : process (reset, clk )
    variable count_flit : integer range 0 to max_no_flit_in_packet := 0;  -- keeps track of the number of flits
    variable zero       : std_logic_vector(flit_size-1 downto 0)   := (others => '0');
  begin  -- process
    if reset = '1' then                 -- asynchronous reset (active high)
      for i in 0 to max_no_flit_in_packet-1 loop
        in_flit_array(i)            <= (others                                => '0');
      end loop;  -- i
      packet_received               <= '0';
      count_flit                                                   := 0;
    elsif falling_edge(clk) then
      if get_check = '1' then
        count_flit                                                 := count_flit + 1;
        if inFlit(flit_size-1) = '0' then
          packet_received           <= '0';
        elsif inFlit(flit_size-1) = '1' then
          packet_received           <= '1';
        end if;
      else
        packet_received             <= '0';
      end if;
      if count_flit = 0 then            -- this only happens when get_check is
                                        -- 0, which is when all channels are empty
        in_flit_array(count_flit)   <= inFlit(flit_size-2 downto 0);
      else
        in_flit_array(count_flit-1) <= inFlit(flit_size-2 downto 0);
        if inFlit(flit_size-1) = '1' then
          count_flit                                               := 0;
        end if;
      end if;
    end if;
  end process Packet_Assemble;


-------------------------------------------------------------------------------
-- Name: Packet Decode
-- Purpose: Given an assembled packet and the priority of the packet, Packet
-- Decode extracts the content of the packet, deciphers the type of packet it
-- is and outputs the content to be send to either Req E2E Unit if the packet
-- is a response packet or Resp E2E if the packet is a request packet. 
-------------------------------------------------------------------------------
  Decode : process (in_flit_array, packet_received, reset)
  begin  -- process
    if reset = '1' then
      packet_decode                         <= (others => '0');
      routepath                             <= (others => '0');
      MAddr                                 <= (others => '0');
      MCmd                                  <= (others => '0');
      MDataInfo                             <= (others => '0');
      MData                                 <= (others => '0');
      MFlag                                 <= (others => '0');
      Request_Valid                         <= '0';
      SResp                                 <= (others => '0');
      SDataInfo                             <= (others => '0');
      SData                                 <= (others => '0');
      Response_Valid                        <= '0';
    end if;
    if packet_received = '1' then
      if inChPri = BE_L or inChPri = BE_H then
        routepath                           <= in_flit_array(0);
        MAddr(15 downto 0)                  <= in_flit_array(1);
        MDataInfo                           <= in_flit_array(2)(11 downto 4);
        MAddr(31 downto 16)                 <= in_flit_array(3);
        if in_flit_array(3)(7 downto 0) = X"FF" and in_flit_array(1)(15 downto 8) = X"FD" then
          if in_flit_array(2)(15 downto 13) = IDLE and in_flit_array(2)(12) = '1' then
            -- this is a Setup Response packet 
            packet_decode                   <= Setup_Resp;
            MCmd                            <= in_flit_array(2)(15 downto 13);
            SResp                           <= in_flit_array(4)(15 downto 14);
            SDataInfo(15 downto 0)          <= in_flit_array(5);  -- to master core
            SDataInfo(31 downto 16)         <= in_flit_array(6);
            SData(15 downto 0)              <= in_flit_array(7);  -- to connID table
            SData(31 downto 16)             <= in_flit_array(8);
            Response_Valid                  <= '1';
            Request_Valid                   <= '0';
          elsif in_flit_array(2)(15 downto 13) = IDLE and in_flit_array(2)(12) = '0' then
            -- this is a Teardown Response packet
            packet_decode                   <= Teardown_Resp;
            MCmd                            <= in_flit_array(2)(15 downto 13);
            SResp                           <= in_flit_array(4)(15 downto 14);
            Response_Valid                  <= '1';
            Request_Valid                   <= '0';
          elsif in_flit_array(2)(15 downto 13) = RD then
            -- this is a Setup Request packet (signals go to Network Controller core.)
            packet_decode                   <= Setup_Req;
            MCmd                            <= in_flit_array(2)(15 downto 13);
            MData(15 downto 0)              <= in_flit_array(4);
            MData(31 downto 16)             <= in_flit_array(5);
            MFlag(15 downto 0)              <= in_flit_array(6);
            MFlag(31 downto 16)             <= in_flit_array(7);
            Request_Valid                   <= '1';
            Response_Valid                  <= '0';
          elsif in_flit_array(2)(15 downto 13) = WR then
            -- this is a Teardown Request packet
            packet_decode                   <= Teardown_Req;
            MCmd                            <= in_flit_array(2)(15 downto 13);
            MFlag(15 downto 0)              <= in_flit_array(4);
            MFlag(31 downto 16)             <= in_flit_array(5);
            Request_Valid                   <= '1';
            Response_Valid                  <= '0';
          end if;
        elsif in_flit_array(1)(15 downto 8) <= X"FC" and in_flit_array(1)(15 downto 8) >= X"00" and in_flit_array(3)(7 downto 0) <= X"FF" and in_flit_array(3)(7 downto 0) >= X"00" then
          if in_flit_array(2)(15 downto 13) = IDLE then
            -- this is BE response packet
            packet_decode                   <= BE_Resp;
            SResp                           <= in_flit_array(4)(15 downto 14);
            if in_flit_array(4)(13) = '1' then
              MCmd                          <= RD;
              SData(15 downto 0)            <= in_flit_array(5);
              SData(31 downto 16)           <= in_flit_array(6);
            else
              MCmd                          <= WR;
            end if;
            Response_Valid                  <= '1';
            Request_Valid                   <= '0';
          else
            -- this is BE request packet
            packet_decode                   <= BE_Req;
            MCmd                            <= in_flit_array(2)(15 downto 13);
            if in_flit_array(2)(15 downto 13) = WR or in_flit_array(2)(15 downto 13) = WRNP or in_flit_array(2)(15 downto 13) = WRC then
              MData(15 downto 0)            <= in_flit_array(4);
              MData(31 downto 16)           <= in_flit_array(5);
              Request_Valid                 <= '1';
              Response_Valid                <= '0';
            end if;
          end if;
        end if;
      elsif inChPri = GP or inChPri = GT then
        if in_flit_array(0)(15 downto 13) = IDLE then
          -- GS Response packet
          packet_decode                     <= GS_Resp;
          SResp                             <= in_flit_array(0)(12 downto 11);
          if in_flit_array(0)(10) = '1' then
            MCmd                            <= RD;
          else
            MCmd                            <= WR;
          end if;
          SData(15 downto 0)                <= in_flit_array(1);
          SData(31 downto 16)               <= in_flit_array(2);
          Response_Valid                    <= '1';
          Request_Valid                     <= '0';
        else
          -- GS Request packet
          packet_decode                     <= GS_Req;
          MCmd                              <= in_flit_array(0)(15 downto 13);
          MFlag(7 downto 0)                 <= in_flit_array(0)(12 downto 5);
          MFlag(23 downto 8)                <= X"FFFD";
          MFlag(31 downto 24)               <= SBA;
          MAddr(15 downto 0)                <= in_flit_array(1);
          MAddr(31 downto 16)               <= in_flit_array(2);
          MData(15 downto 0)                <= in_flit_array(3);
          MData(31 downto 16)               <= in_flit_array(4);
          Request_Valid                     <= '1';
          Response_Valid                    <= '0';
        end if;
      end if;
    end if;
  end process;

-------------------------------------------------------------------------------
-- Name: Signal Validation
-- Purpose: Sends a request or response validation signal to either ReqE2E or
-- ResE2E as pulse handshake signals so that those two components can
-- recognizes the arrival of new data. 
-------------------------------------------------------------------------------
  Signal_Validation : process (clk, reset)
  begin  -- process
    if reset = '1' then                 -- asynchronous reset (active low)
      Response_arrived   <= '0';
      Request_arrived    <= '0';
    elsif clk'event and clk = '1' then  -- rising clock edge
      if packet_received = '1' and Response_Valid = '1' then
        Response_arrived <= '1';
        Request_arrived  <= '0';
      elsif packet_received = '1' and Request_Valid = '1' then
        Request_arrived  <= '1';
        Response_arrived <= '0';
      else
        Response_arrived <= '0';
        Request_arrived  <= '0';
      end if;
    end if;
  end process;

end behavioral;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;

package Decap_pkg is
  component Decap
    port (
      clk              : in  std_logic;
      reset            : in  std_logic;
      SBA              : in  std_logic_vector(SBA_length-1 downto 0);
      empty            : in  std_logic_vector(num_vc-1 downto 0);
      get              : out std_logic_vector(num_vc-1 downto 0);
      inChSel          : out std_logic_vector(chSelect-1 downto 0);
      inFlit           : in  std_logic_vector(flit_size-1 downto 0);
      routepath        : out std_logic_vector(routepath_length-1 downto 0);
      MAddr            : out std_logic_vector(MAddr_length-1 downto 0);
      MCmd             : out std_logic_vector(MCmd_length-1 downto 0);
      packet_decode    : out Packet_Type;
      MFlag            : out std_logic_vector(MFlag_length-1 downto 0);
      MData            : out std_logic_vector(MData_length-1 downto 0);
      MDataInfo        : out std_logic_vector(MDataInfo_length-1 downto 0);
      Request_arrived  : out std_logic;
      SResp            : out std_logic_vector(SResp_length-1 downto 0);
      SData            : out std_logic_vector(SData_length-1 downto 0);
      SDataInfo        : out std_logic_vector(SDataInfo_length-1 downto 0);
      Response_arrived : out std_logic);
  end component;
end Decap_pkg;


