-------------------------------------------------------------------------------
-- Title      : Encapsulation Unit
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : Encap.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/05
-------------------------------------------------------------------------------
-- Description: The Encapsulation Unit receives requests from the ReqE2E unit.
-- It packetizes the request according to the DTU NoC packet format and sends
-- the packet out to one of the specified outgoing virtual channels queues if
-- the packet is a GS packet, or schedule a BE virtual channel to send if the
-- packet is a BE packet. If the specified channel queue is full, the Encapsulation Unit
-- will withhold its acknowledge to the ReqE2E unit. Therefore, stopping it
-- from receiving new requests. 
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;

entity Encap is
  port (
    clk              : in  std_logic;
    reset            : in  std_logic;
    -- Signal from ReqE2E
    packet_encode_In : in  Packet_Type;
    MCmd_In          : in  MCmd_type;
    MAddr_In         : in  std_logic_vector(MAddr_length-1 downto 0);
    MFlag_In         : in  std_logic_vector(MFlag_length-1 downto 0);
    MData_In         : in  std_logic_vector(MData_length-1 downto 0);
    MDataInfo_In     : in  std_logic_vector(MDataInfo_length-1 downto 0);
    request_phase    : in  std_logic;
    -- Signal to ReqE2E
    packet_sent_Out  : out std_logic;
    -- Signal from Connection ID Table
    InjID            : in  std_logic_vector(31 downto 0);
    -- Signal from Route Lookup Table
    routepath_In     : in  std_logic_vector(routepath_length-1 downto 0);
    -- Signal to and from Network Interface
    outChSel         : out std_logic_vector(chSelect-1 downto 0);
    packet           : out outflitRegType;
    hold             : in  std_logic_vector(7 downto 0)
    );
end Encap;

architecture behavioral of Encap is

  signal regs           : RegType;
  signal out_flit_array : outflitRegType;

  type req_state_type is (INIT, IDLE, PACKETIZE, SEND, AWAIT);
  signal cs, ns : req_state_type;

begin  -- behavioral

  FSM_register : process (clk, reset)
  begin  -- process FSM
    if reset = '1' then                 -- asynchronous reset (active high)
      cs <= INIT;
    elsif clk'event and clk = '1' then  -- rising clock edge
      cs <= ns;
    end if;
  end process FSM_register;

  FSM_combinatorial               : process (cs, request_phase, hold, packet_encode_In, MCmd_In, MAddr_In, MFlag_In, MData_In, MDataInfo_In, routepath_In, InjID)
    variable   ch_selected        : integer range 0 to num_vc-1;
    variable   channels_available : std_logic_vector(7 downto 0)                      := (others => '0');
    function to_int(X             : std_logic_vector) return integer is
      variable result             : integer;
    begin
      result                                                                          := 0;
      for i in X'range loop
        result                                                                        := result * 2;
        case X(i) is
          when '0' | 'L'                                                    => null;
          when '1' | 'H'                                                    => result := result + 1;
          when others                                                       => null;
        end case;
      end loop;
      return result;
    end function to_int;
  begin  -- process FSM_next
    ns                      <= cs;
    case cs is
      when INIT                                                             =>
        for i in 0 to max_no_flit_in_packet-1 loop
          out_flit_array(i) <= (others                                      => '0');
          packet(i)         <= (others                                      => '0');
        end loop;  -- i
        packet_sent_Out     <= '0';
        outChSel            <= (others                                      => '0');
        ns                  <= IDLE;
      when IDLE                                                             =>
        if request_phase = '1' then
          ns                <= PACKETIZE;
        else
          ns                <= IDLE;
        end if;
        packet_sent_Out     <= '0';

      when PACKETIZE                                =>
        packet_sent_Out         <= '0';
        case packet_encode_In is
          when BE_Req                               =>
            out_flit_array(0)   <= '0' & routepath_In;
            out_flit_array(1)   <= '0' & MAddr_In(15 downto 0);
            out_flit_array(2)   <= '0' & MCmd_In & '0' & MDataInfo_In & "0000";
            if MCmd_In = RD or MCmd_In = RDEX or MCmd_In = RDL then
              out_flit_array(3) <= '1' & MAddr_In(31 downto 16);
            elsif MCmd_In = WR or MCmd_In = WRNP or MCmd_In = WRC or MCmd_In = BCST then
              out_flit_array(3) <= '0' & MAddr_In(31 downto 16);
              out_flit_array(4) <= '0' & MData_In(15 downto 0);
              out_flit_array(5) <= '1' & MData_In(31 downto 16);
            end if;
          when Setup_Req                            =>
            out_flit_array(0)   <= '0' & routepath_In;
            out_flit_array(1)   <= '0' & MAddr_In(15 downto 0);
            out_flit_array(2)   <= '0' & MCmd_In & '0' & MDataInfo_In & "0000";
            out_flit_array(3)   <= '0' & MAddr_In(31 downto 16);
            out_flit_array(4)   <= '0' & MData_In(15 downto 0);
            out_flit_array(5)   <= '0' & MData_In(31 downto 16);
            out_flit_array(6)   <= '0' & MFlag_In(15 downto 0);
            out_flit_array(7)   <= '1' & MFlag_In(31 downto 16);
          when GS_Req                               =>
            out_flit_array(0)   <= '0' & MCmd_In & MFlag_In(7 downto 0) & "00000";
            out_flit_array(1)   <= '0' & MAddr_In(15 downto 0);
            if MCmd_In = RD or MCmd_In = RDEX or MCmd_In = RDL then
              out_flit_array(2) <= '1' & MAddr_In(31 downto 16);
            elsif MCmd_In = WR or MCmd_In = WRNP or MCmd_In = WRC or MCmd_In = BCST then
              out_flit_array(2) <= '0' & MAddr_In(31 downto 16);
              out_flit_array(3) <= '0' & MData_In(15 downto 0);
              out_flit_array(4) <= '1' & MData_In(31 downto 16);
            end if;
          when Teardown_Req                         =>
            out_flit_array(0)   <= '0' & routepath_In;
            out_flit_array(1)   <= '0' & MAddr_In(15 downto 0);
            out_flit_array(2)   <= '0' & MCmd_In & '0' & MDataInfo_In & "0000";
            out_flit_array(3)   <= '0' & MAddr_In(31 downto 16);
            out_flit_array(4)   <= '0' & MFlag_In(15 downto 0);
            out_flit_array(5)   <= '1' & MFlag_In(31 downto 16);
          when others                               => null;
        end case;
        if packet_encode_In /= GS_Req then
          channels_available             := (others => '0');
          for ch_num in 0 to num_vc-1 loop
            if regs(ch_num) = BE_L and hold(ch_num) = '0' then
              channels_available(ch_num) := '1';
            end if;
          end loop;  -- i
          for i in 0 to 7 loop
            if channels_available(i) = '1' then
              ch_selected                := i;
              exit;
            end if;
          end loop;  -- i             
        else
          ch_selected                    := to_int(InjID(7 downto 4));
        end if;
        if hold(ch_selected) = '0' then
          ns                    <= SEND;
        elsif hold(ch_selected) = '1' then
          ns                    <= AWAIT;
        end if;
      when SEND                                     =>
        ns                      <= IDLE;
        outChSel                <= std_logic_vector(to_unsigned(ch_selected, chSelect));
        packet                  <= out_flit_array;
        packet_sent_Out         <= '1';

      when AWAIT =>
        if hold(ch_selected) = '0' then
          ns            <= SEND;
        else
          ns            <= AWAIT;
        end if;
        packet_sent_Out <= '0';

      when others => null;
    end case;
  end process FSM_combinatorial;

-------------------------------------------------------------------------------
-- Name: Outgoing Virtual Channel Priority Table
-- Purpose: Allow the priority scheduler to know the priority of each of the
-- incoming virtual channels. This table is currently set statically. 
-------------------------------------------------------------------------------

  regs(0) <= BE_L;
  regs(1) <= BE_L;
  regs(2) <= BE_L;
  regs(3) <= BE_L;
  regs(4) <= GP;
  regs(5) <= GP;
  regs(6) <= GT;
  regs(7) <= GT;


end behavioral;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;

package Encap_pkg is
  component Encap
    port (
      clk              : in  std_logic;
      reset            : in  std_logic;
      packet_encode_In : in  Packet_Type;
      MCmd_In          : in  MCmd_type;
      MAddr_In         : in  std_logic_vector(MAddr_length-1 downto 0);
      MFlag_In         : in  std_logic_vector(MFlag_length-1 downto 0);
      MData_In         : in  std_logic_vector(MData_length-1 downto 0);
      MDataInfo_In     : in  std_logic_vector(MDataInfo_length-1 downto 0);
      request_phase    : in  std_logic;
      packet_sent_Out  : out std_logic;
      InjID            : in  std_logic_vector(31 downto 0);
      routepath_In     : in  std_logic_vector(routepath_length-1 downto 0);
      outChSel         : out std_logic_vector(chSelect-1 downto 0);
      packet           : out outflitRegType;
      hold             : in  std_logic_vector(7 downto 0));
  end component;
end Encap_pkg;
