-------------------------------------------------------------------------------
-- Title      : Output Queue
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : Queue.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/05
-------------------------------------------------------------------------------
-- Description: Thehe output queues are used to hold
-- packets which are intended to be sent to the queue's corresponding
-- virtual channel.
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;


entity Queue is
  port (
    clk        : in  std_logic;
    reset      : in  std_logic;
    load       : in  std_logic;
    flit_array : in  outflitRegType;
    outFlit    : out std_logic_vector(flit_size-1 downto 0);
    ready      : out std_logic;
    full       : in  std_logic;
    put        : out std_logic);
end Queue;

architecture behavorial of Queue is
  signal flit_counter : integer range 0 to max_no_flit_in_packet-1 := 0;
  signal queue        : outflitRegType;
  signal last_flit    : boolean                                    := false;
begin  -- behavorial

  send              : process (clk, reset)
    variable loaded : boolean := false;
  begin  -- process send
    if reset = '1' then                 -- asynchronous reset (active low)
      flit_counter     <= 0;
      put              <= '0';
      ready            <= '1';          -- ready = 1 when the queue is empty
      outFlit          <= (others => '0');
      loaded                  := false;
      last_flit        <= false;
      for i in 0 to max_no_flit_in_packet-1 loop
        queue(i)       <= (others => '0');
      end loop;  -- i
    elsif clk'event and clk = '1' then  -- rising clock edge
      if load = '1' then
        ready          <= '0';
        queue          <= flit_array;
        loaded                := true;
        last_flit      <= false;
      elsif full = '0' and loaded = true then
        if queue(flit_counter)(16) = '0' then
          outFlit      <= queue(flit_counter);
          put          <= '1';
          flit_counter <= flit_counter + 1;
          last_flit    <= false;
        elsif queue(flit_counter)(16) = '1' then
          outFlit      <= queue(flit_counter);
          put          <= '1';
          flit_counter <= 0;
          loaded := false;
          last_flit    <= true;
        end if;
      elsif full = '1' and loaded = true then
        put            <= '0';
        ready          <= '0';
      end if;
      if last_flit = true then
        ready          <= '1';
        put            <= '0';
        last_flit      <= false;
      end if;
    end if;
  end process send;
end behavorial;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;
package Queue_pkg is
  component Queue
    port (
      clk        : in  std_logic;
      reset      : in  std_logic;
      load       : in  std_logic;
      flit_array : in  outflitRegType;
      outFlit    : out std_logic_vector(flit_size-1 downto 0);
      ready      : out std_logic;
      full       : in  std_logic;
      put        : out std_logic);
  end component;
end Queue_pkg;
