-------------------------------------------------------------------------------
-- Title      : Queue Control Unit
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : Queue_Control.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/04
-------------------------------------------------------------------------------
-- Description: Controls the loading and enabling of all 8 output queues, and
-- communicates the queque status to the Encapsulation Unit. 
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;

entity Queue_Control is

  port (
    clk         : in  std_logic;
    reset       : in  std_logic;
    load        : out std_logic_vector(7 downto 0);
    ready       : in  std_logic_vector(7 downto 0);
    outChSel    : in  std_logic_vector(2 downto 0);
    packet_sent : in  std_logic;
    hold        : out std_logic_vector(7 downto 0));
end Queue_Control;

architecture behavioral of Queue_Control is
  type req_state_type is (INIT, IDLE, REQUEST_RECEIVED, LOAD_QUEUE, AWAIT);
  signal cs_req, ns_req : req_state_type;

begin  -- behavioral 

  Req_register : process (clk, reset)
  begin  -- process FSM
    if reset = '1' then                 -- asynchronous reset (active high)
      cs_req <= INIT;
    elsif clk'event and clk = '1' then  -- rising clock edge
      cs_req <= ns_req;
    end if;
  end process Req_register;

  Req_combinatorial         : process (cs_req, packet_sent, outChSel, ready)
    variable   outChSel_int : integer;
    function to_int(X       : std_logic_vector) return integer is
      variable result       : integer;
    begin
      result                                    := 0;
      for i in X'range loop
        result                                  := result * 2;
        case X(i) is
          when '0' | 'L'              => null;
          when '1' | 'H'              => result := result + 1;
          when others                 => null;
        end case;
      end loop;
      return result;
    end function to_int;
  begin  -- process FSM_next
    ns_req                 <= cs_req;
    case cs_req is
      when INIT                       =>
        hold               <= (others => '0');
        load               <= (others => '0');
        ns_req             <= IDLE;
      when IDLE                       =>
        if packet_sent = '1' then
          outChSel_int                          := to_int(outChSel);
          if ready(outChSel_int) = '1' then
            ns_req         <= LOAD_QUEUE;
          else
            ns_req         <= AWAIT;
          end if;
        else
          ns_req           <= IDLE;
        end if;
        load               <= (others => '0');
        for i in 0 to 7 loop
          if ready(i) = '1' then
            hold(i)        <= '0';
          end if;
        end loop;  -- i
      when LOAD_QUEUE                 =>
        for i in 0 to 7 loop
          if i = outChSel_int then
            load(i)        <= '1';
          else
            load(i)        <= '0';
          end if;
        end loop;  -- i
        hold(outChSel_int) <= '1';
        ns_req             <= IDLE;
        for i in 0 to 7 loop
          if ready(i) = '1' and i /= outChSel_int then
            hold(i)        <= '0';
          end if;
        end loop;  -- i
      when AWAIT                      =>
        if ready(outChSel_int) = '1' then
          ns_req           <= LOAD_QUEUE;
        else
          ns_req           <= AWAIT;
        end if;
        load               <= (others => '0');
        hold(outChSel_int) <= '1';
        for i in 0 to 7 loop
          if ready(i) = '1' and i /= outChSel_int then
            hold(i)        <= '0';
          end if;
        end loop;  -- i
      when others                     => null;
    end case;
  end process Req_combinatorial;
end behavioral;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;

package Queue_Control_pkg is
  component Queue_Control
    port (
      clk         : in  std_logic;
      reset       : in  std_logic;
      load        : out std_logic_vector(7 downto 0);
      ready       : in  std_logic_vector(7 downto 0);
      packet_sent : in  std_logic;
      outChSel    : in  std_logic_vector(2 downto 0);
      hold        : out std_logic_vector(7 downto 0));
  end component;
end Queue_Control_pkg;
