-------------------------------------------------------------------------------
-- Title      : Request End-to-End Flow Control Unit
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : ReqE2E.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/16
-------------------------------------------------------------------------------
-- Description: The ReqE2E unit accepts requests from the master core,
-- determines the type of request it is, and sends the necessary information to
-- the Encap Unit for packaging.
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;


entity ReqE2E is
  port (
    clk              : in  std_logic;
    reset            : in  std_logic;
    -- OCP Request Phase Signals
    MCmd             : in  std_logic_vector(2 downto 0);
    MAddr            : in  std_logic_vector(MAddr_length-1 downto 0);
    MData            : in  std_logic_vector(MData_length-1 downto 0);
    MFlag            : in  std_logic_vector(MFlag_length-1 downto 0);
    MReqInfo         : in  std_logic_vector(MReqInfo_length-1 downto 0);
    MDataInfo        : in  std_logic_vector(MDataInfo_length-1 downto 0);
    SCmdAccept       : out std_logic;
    -- OCP Response Phase Signals
    SResp            : out SResp_type;
    SData            : out std_logic_vector(SData_length-1 downto 0);
    SDataInfo        : out std_logic_vector(SDataInfo_length-1 downto 0);
    MRespAccept      : in  std_logic;
    -- Signals to Encap
    packet_encode    : out Packet_Type;
    MAddr_Out        : out std_logic_vector(MAddr_length-1 downto 0);
    MCmd_Out         : out MCmd_type;
    MFlag_Out        : out std_logic_vector(MFlag_length-1 downto 0);
    MData_Out        : out std_logic_vector(MData_length-1 downto 0);
    MDataInfo_Out    : out std_logic_vector(MDataInfo_length-1 downto 0);
    request_phase    : out std_logic;
    -- Signals from Encap
    packet_sent      : in  std_logic;
    -- Signals from Decap
    packet_decode_In : in  Packet_Type;
    MCmd_In          : in  MCmd_type;   -- to decipher between rd and wr response
    -- for setup response packet, MAddr carries connID,which will be mapped to
    -- SData to be sent to the master core.
    MAddr_In         : in  std_logic_vector(MAddr_length-1 downto 0);
    SResp_In         : in  SResp_type;
    SData_In         : in  std_logic_vector(SData_length-1 downto 0);
    SDataInfo_In     : in  std_logic_vector(SDataInfo_length-1 downto 0);
    Response_arrived : in  std_logic;
    -- Signal to Connection ID Table
    read_address     : out std_logic_vector(31 downto 0);
    write_address    : out std_logic_vector(31 downto 0);
    write_data       : out std_logic_vector(31 downto 0);
    MemWrite         : out std_logic;
    MemRead          : out std_logic;
    -- Signal to Route Lookup Table
    Dest_SBA         : out std_logic_vector(7 downto 0));
end ReqE2E;

architecture behavioral of ReqE2E is
-- Signal Declaration
  type req_state_type is (IDLE, REQUEST_RECEIVED, PACKETIZED);
  signal cs_req, ns_req      : req_state_type;
  type resp_state_type is (IDLE, RESPONSE_RECEIVED);
  signal cs_resp, ns_resp    : resp_state_type;
  signal response_phase      : std_logic;
  signal request_encode      : Packet_Type;
  signal addr_req, addr_resp : std_logic_vector(31 downto 0);


begin  -- behavioral

-------------------------------------------------------------------------------
-- Name: Request Finite State Machine
-- Purpose: Controls the OCP interface at the request receiving end.
-------------------------------------------------------------------------------
  Req_register : process (clk, reset)
  begin  -- process FSM
    if reset = '1' then                 -- asynchronous reset (active high)
      cs_req <= IDLE;
    elsif clk'event and clk = '1' then  -- rising clock edge
      cs_req <= ns_req;
    end if;
  end process Req_register;

  Req_combinatorial : process (cs_req, MCmd, packet_sent)

  begin  -- process FSM_next
    ns_req            <= cs_req;
    case cs_req is
      when IDLE             =>
        if MCmd /= "000" then
          ns_req      <= REQUEST_RECEIVED;
        else
          ns_req      <= IDLE;
        end if;
        SCmdAccept    <= '0';
        request_phase <= '0';
      when REQUEST_RECEIVED =>
        if packet_sent = '1' then
          ns_req      <= PACKETIZED;
        else
          ns_req      <= REQUEST_RECEIVED;
        end if;
        SCmdAccept    <= '0';
        request_phase <= '1';
      when PACKETIZED       =>
        ns_req        <= IDLE;
        SCmdAccept    <= '1';
        request_phase <= '0';
      when others           => null;
    end case;
  end process Req_combinatorial;

-------------------------------------------------------------------------------
-- Name: Response Finite State Machine
-- Purpose: Controls the OCP interface signals at the response end.
-------------------------------------------------------------------------------
  Resp_register : process (clk, reset)
  begin  -- process FSM
    if reset = '1' then                 -- asynchronous reset (active high)
      cs_resp <= IDLE;
    elsif clk'event and clk = '1' then  -- rising clock edge
      cs_resp <= ns_resp;
    end if;
  end process Resp_register;

  Resp_combinatorial : process (cs_resp, Response_arrived, packet_decode_In, MCmd_In, MAddr_In, SResp_In, SData_In, SDataInfo_In)
  begin  -- process FSM_next
    ns_resp                 <= cs_resp;
    write_address               <= (others => '0');
    write_data              <= (others => '0');
    MemWrite                <= '0';
    case cs_resp is
      when IDLE                        =>
        if Response_arrived = '1' then
          ns_resp           <= RESPONSE_RECEIVED;
        else
          ns_resp           <= IDLE;
        end if;
        SResp               <= NV;
        SData               <= (others => '0');
        SDataInfo           <= (others => '0');
        response_phase      <= '0';
      when RESPONSE_RECEIVED           =>
        ns_resp             <= IDLE;
        response_phase      <= '1';
        case packet_decode_In is
          when Setup_Resp              =>
            SResp           <= SResp_In;
            if SResp_In = DVA then
              SData         <= MAddr_In;  -- assigning the connectionID to SData
              SDataInfo     <= SDataInfo_In;  -- contain the address of dest core
                                        -- connection was set up to.
              write_address <= MAddr_In;
              write_data    <= SData_In;
              MemWrite      <= '1';
            else
              SData         <= (others => '0');
              SDataInfo     <= (others => '0');
              MemWrite      <= '0';
            end if;

          when Teardown_Resp           =>
            SResp           <= SResp_In;
            if SResp_In = DVA then
              write_address <= MAddr_In;
              write_data    <= (others => '0');
              MemWrite      <= '1';
            end if;
            SData           <= MAddr_In;  -- assgined the connectionID to SData,
            SDataInfo       <= SDataInfo_In;  -- to indicate to the core that this
                                        -- connection was successfully torn down.
          when BE_Resp                 =>
            SResp           <= SResp_In;
            if MCmd_In = RD then
              SData         <= SData_In;
            elsif MCmd_In = WR then
              SData         <= (others => '0');
            end if;
            SDataInfo       <= SDataInfo_In;
          when GS_Resp                 =>
            SResp           <= SResp_In;
            if MCmd_In = RD then
              SData         <= SData_In;
            elsif MCmd_In = WR then
              SData         <= (others => '0');
            end if;
            SDataInfo       <= SDataInfo_In;
          when others                  => null;
        end case;
      when others                      => null;
    end case;
  end process Resp_combinatorial;

-------------------------------------------------------------------------------
-- Name: Request Type
-- Purpose:Decide what type of request the core requested.
-- Possible types can be: GS Setup, GS Teardown, GS Use, and BE Use
-------------------------------------------------------------------------------
  Request_Type : process (clk, reset)
  begin  -- process
    if reset = '1' then                 -- asynchronous reset (active low)
      read_address   <= (others => '0');
      MemRead        <= '0';
      request_encode <= (others => '0');

    elsif clk'event and clk = '1' then  -- rising clock edge
      if MReqInfo = "00" then
        -- request is a general best effort read/write request
        request_encode <= BE_Req;
        MemRead        <= '0';
      elsif MReqInfo = "01" and MCmd = "010" then
        -- request is for a guaranteed service setup request
        request_encode <= Setup_Req;
        MemRead        <= '0';
      elsif MReqInfo = "10" then
        -- request is a guaranteed service use read/write request
        request_encode <= GS_Req;
        read_address   <= MFlag;
        MemRead        <= '1';
      elsif MReqInfo = "11" and MCmd = "001" then
        -- request is a guaranteed service teardown request
        request_encode <= Teardown_Req;
        MemRead        <= '0';
      end if;
    end if;
  end process Request_Type;

  packet_encode <= request_encode;


-------------------------------------------------------------------------------
-- Name: To Encap
-- Purpose: Based on the request type, determines the message to send to the
-- Encap unit for packaging.
-------------------------------------------------------------------------------

  To_encap : process(request_encode)
  begin  -- process To_encap
    MDataInfo_Out <= MDataInfo;
    MCmd_Out      <= MCmd;
    MAddr_Out     <= MAddr;

    if request_encode = BE_Req then
      MData_Out <= MData;
      MFlag_Out <= (others => '0');
      Dest_SBA  <= MAddr(31 downto 24);
    elsif request_encode = Setup_Req then
      MData_Out <= MData;
      MFlag_Out <= MFlag;
      Dest_SBA  <= NC_SBA;
    elsif request_encode = Teardown_Req then
      MFlag_Out <= MFlag;
      MData_Out <= (others => '0');
      Dest_SBA  <= NC_SBA;
    elsif request_encode = GS_Req then
      MData_Out <= MData;
      MFlag_Out <= MFlag;
      Dest_SBA  <= (others => '0');
    end if;
  end process To_encap;

end behavioral;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;

package ReqE2E_pkg is
  component ReqE2E
    port (
      clk              : in  std_logic;
      reset            : in  std_logic;
      MCmd             : in  std_logic_vector(2 downto 0);
      MAddr            : in  std_logic_vector(MAddr_length-1 downto 0);
      MData            : in  std_logic_vector(MData_length-1 downto 0);
      MFlag            : in  std_logic_vector(MFlag_length-1 downto 0);
      MReqInfo         : in  std_logic_vector(MReqInfo_length-1 downto 0);
      MDataInfo        : in  std_logic_vector(MDataInfo_length-1 downto 0);
      SCmdAccept       : out std_logic;
      SResp            : out SResp_type;
      SData            : out std_logic_vector(SData_length-1 downto 0);
      SDataInfo        : out std_logic_vector(SDataInfo_length-1 downto 0);
      MRespAccept      : in  std_logic;
      packet_encode    : out Packet_Type;
      MAddr_Out        : out std_logic_vector(MAddr_length-1 downto 0);
      MCmd_Out         : out MCmd_type;
      MFlag_Out        : out std_logic_vector(MFlag_length-1 downto 0);
      MData_Out        : out std_logic_vector(MData_length-1 downto 0);
      MDataInfo_Out    : out std_logic_vector(MDataInfo_length-1 downto 0);
      request_phase    : out std_logic;
      packet_sent      : in  std_logic;
      packet_decode_In : in  Packet_Type;
      MCmd_In          : in  MCmd_type;
      MAddr_In         : in  std_logic_vector(MAddr_length-1 downto 0);
      SResp_In         : in  SResp_type;
      SData_In         : in  std_logic_vector(SData_length-1 downto 0);
      SDataInfo_In     : in  std_logic_vector(SDataInfo_length-1 downto 0);
      Response_arrived : in  std_logic;
      read_address     : out std_logic_vector(31 downto 0);
      write_address    : out std_logic_vector(31 downto 0);
      write_data       : out std_logic_vector(31 downto 0);
      MemWrite         : out std_logic;
      MemRead          : out std_logic;
      Dest_SBA         : out std_logic_vector(7 downto 0));
  end component;
end ReqE2E_pkg;
