-------------------------------------------------------------------------------
-- Title      : Slave Network Adatper
-- Project    : OCP compliant Network Adapter
-------------------------------------------------------------------------------
-- File       : Slave.vhd
-- Author     : Juliana Pei Zhou
-- Company    : Informatics and Mathematical Modelling, TUD
-- Last update: 2004/09/16
-------------------------------------------------------------------------------
-- Description: The top level behavior code for the slave network adapter.
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;
use work.Decap_pkg.all;
use work.Encap_pkg.all;
use work.ConnIDTable_pkg.all;
use work.RouteLookupTable_pkg.all;
use work.ReqE2E_pkg.all;
use work.Queue_pkg.all;
use work.Queue_Control_pkg.all;


entity NASlave is
  port(
    clk         : in  std_logic;
    reset       : in  std_logic;
    -- OCP Interface signals
    MCmd        : in  std_logic_vector(MCmd_length-1 downto 0);
    MAddr       : in  std_logic_vector(MAddr_length-1 downto 0);
    MData       : in  std_logic_vector(MData_length-1 downto 0);
    MFlag       : in  std_logic_vector(MFlag_length-1 downto 0);
    MReqInfo    : in  std_logic_vector(MReqInfo_length-1 downto 0);
    MDataInfo   : in  std_logic_vector(MDataInfo_length-1 downto 0);
    SCmdAccept  : out std_logic;
    SResp       : out std_logic_vector(SResp_length-1 downto 0);
    SData       : out std_logic_vector(SData_length-1 downto 0);
    SDataInfo   : out std_logic_vector(SDataInfo_length-1 downto 0);
    MRespAccept : in  std_logic;
    -- Network Interface signals
    empty       : in  std_logic_vector(num_vc-1 downto 0);
    get         : out std_logic_vector(num_vc-1 downto 0);
    inChannels  : in  channel_type;
    full        : in  std_logic_vector(num_vc-1 downto 0);
    put         : out std_logic_vector(num_vc-1 downto 0);
    outChannels : out channel_type
    );
end entity NASlave;

architecture behavioral of NASlave is
-- signal declarations

  -- ReqE2E (to Encap)
  signal packet_encode_i : Packet_Type;
  signal MAddr_Out_i     : std_logic_vector(MAddr_length-1 downto 0);
  signal MCmd_Out_i      : MCmd_type;
  signal MFlag_Out_i     : std_logic_vector(MFlag_length-1 downto 0);
  signal MData_Out_i     : std_logic_vector(MData_length-1 downto 0);
  signal MDataInfo_Out_i : std_logic_vector(MDataInfo_length-1 downto 0);
  signal request_phase_i : std_logic;
  -- ReqE2E (from Encap)
  signal packet_sent_i   : std_logic;

  -- ReqE2E (to Connection ID Table)
  signal req_write_address_i       : std_logic_vector(31 downto 0);
  signal req_write_data_i   : std_logic_vector(31 downto 0);
  signal req_read_address_i       : std_logic_vector(31 downto 0);
  signal connIDTable_write_en : std_logic;
  signal connIDTable_read_en  : std_logic;

  -- ConnID Table (to Encap)
  signal InjID_i : std_logic_vector(31 downto 0);

  -- ReqE2E (to Route Looup Table)
  signal Dest_SBA_i : std_logic_vector(7 downto 0);

  -- Route Lookup Table (to Encap)
  signal routepath_out_i : std_logic_vector(routepath_length-1 downto 0);

  -- Decap (to ReqE2E)
  signal SResp_i            : std_logic_vector(SResp_length-1 downto 0);
  signal SData_i            : std_logic_vector(SData_length-1 downto 0);
  signal SDataInfo_i        : std_logic_vector(SDataInfo_length-1 downto 0);
  signal MAddr_i            : std_logic_vector(MAddr_length-1 downto 0);
  signal MCmd_i             : std_logic_vector(MCmd_length-1 downto 0);
  signal packet_decode_i    : Packet_Type;
  signal Response_arrived_i : std_logic;


  -- Decap (from input mux)
  signal inFlit_i  : std_logic_vector(flit_size-1 downto 0);
  -- Decap (to input mux)
  signal inChSel_i : std_logic_vector(chSelect-1 downto 0);

  -- Decap (to RespE2E)
  signal routepath_in_i    : std_logic_vector(routepath_length-1 downto 0);
  signal MFlag_i           : std_logic_vector(MFlag_length-1 downto 0);
  signal MData_i           : std_logic_vector(MData_length-1 downto 0);
  signal MDataInfo_i       : std_logic_vector(MDataInfo_length-1 downto 0);
  signal Request_arrived_i : std_logic;

  -- Encap (to Queue Control)
  signal outChSel_i : std_logic_vector(chSelect-1 downto 0);

  -- Encap (from Queue Control)
  signal hold_i : std_logic_vector(7 downto 0);

  -- Encap (to Queue)
  signal packet_i : outflitRegType;

  -- Queue (to Queue Control)
  signal ready : std_logic_vector(7 downto 0);

  -- Queue (from Queue Control)
  signal load : std_logic_vector(7 downto 0);


begin

  ReqE2E_Unit: component ReqE2E
    port map (
        clk              => clk,
        reset            => reset,
        MCmd             => MCmd,
        MAddr            => MAddr,
        MData            => MData,
        MFlag            => MFlag,
        MReqInfo         => MReqInfo,
        MDataInfo        => MDataInfo,
        SCmdAccept       => SCmdAccept,
        SResp            => SResp,
        SData            => SData,
        SDataInfo        => SDataInfo,
        MRespAccept      => MRespAccept,
        packet_encode    => packet_encode_i,
        MAddr_Out        => MAddr_Out_i,
        MCmd_Out         => MCmd_Out_i,
        MFlag_Out        => MFlag_Out_i,
        MData_Out        => MData_Out_i,
        MDataInfo_Out    => MDataInfo_Out_i,
        request_phase    => request_phase_i,
        packet_sent      => packet_sent_i,
        packet_decode_In => packet_decode_i,
        MCmd_In          => MCmd_i,
        MAddr_In         => MAddr_i,
        SResp_In         => SResp_i,
        SData_In         => SData_i,
        SDataInfo_In     => SDataInfo_i,
        Response_arrived => Response_arrived_i,
        read_address     => req_read_address_i,
        write_address    => req_write_address_i,
        write_data       => req_write_data_i,
        MemWrite         => connIDTable_write_en,
        MemRead          => connIDTable_read_en,
        Dest_SBA         => Dest_SBA_i);


  RouteLookupTable_Unit : component RouteLookupTable
    port map (
      address   => Dest_SBA_i,
      read_data => routepath_out_i);

  ConnIDTable_Unit: component ConnIDTable
    port map (
        clk           => clk,
        reset         => reset,
        read_address  => req_read_address_i,
        read_data     => InjID_i,
        write_address => req_write_address_i,
        write_data    => req_write_data_i,
        MemWrite      => connIDTable_write_en,
        MemRead       => connIDTable_read_en);


  Decapsulation : component Decap
    port map (
      clk              => clk,
      reset            => reset,
      SBA              => Master_SBA,
      empty            => empty,
      get              => get,
      inChSel          => inChSel_i,
      inFlit           => inFlit_i,
      routepath        => routepath_in_i,
      MAddr            => MAddr_i,
      MCmd             => MCmd_i,
      packet_decode    => packet_decode_i,
      MFlag            => MFlag_i,
      MData            => MData_i,
      MDataInfo        => MDataInfo_i,
      Request_arrived  => Request_arrived_i,
      SResp            => SResp_i,
      SData            => SData_i,
      SDataInfo        => SDataInfo_i,
      Response_arrived => Response_arrived_i);



  Input_PortMux : process (reset, inChannels, inChsel_i)
  begin  -- process Input Port Mux
    if reset = '1' then
      inFlit_i     <= (others => '0');
    else
      case inChSel_i is
        when "000"            =>
          inFlit_i <= inChannels(0);
        when "001"            =>
          inFlit_i <= inChannels(1);
        when "010"            =>
          inFlit_i <= inChannels(2);
        when "011"            =>
          inFlit_i <= inChannels(3);
        when "100"            =>
          inFlit_i <= inChannels(4);
        when "101"            =>
          inFlit_i <= inChannels(5);
        when "110"            =>
          inFlit_i <= inChannels(6);
        when "111"            =>
          inFlit_i <= inChannels(7);
        when others           => null;
      end case;
    end if;
  end process Input_PortMux;


  Encap_Unit : Encap
    port map (
      clk              => clk,
      reset            => reset,
      packet_encode_In => packet_encode_i,
      MCmd_In          => MCmd_Out_i,
      MAddr_In         => MAddr_Out_i,
      MFlag_In         => MFlag_Out_i,
      MData_In         => MData_Out_i,
      MDataInfo_In     => MDataInfo_Out_i,
      request_phase    => request_phase_i,
      packet_sent_Out  => packet_sent_i,
      InjID            => InjID_i,
      routepath_In     => routepath_out_i,
      outChSel         => outChSel_i,
      packet           => packet_i,
      hold             => hold_i);

  Queue_control_Unit: Queue_Control
    port map (
        clk         => clk,
        reset       => reset,
        load        => load,
        ready       => ready,
        outChSel    => outChSel_i,
        packet_sent => packet_sent_i,
        hold        => hold_i);

  Queue_zero : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(0),
      flit_array => packet_i,
      outFlit    => outChannels(0),
      ready      => ready(0),
      full       => full(0),
      put        => put(0));

  Queue_one : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(1),
      flit_array => packet_i,
      outFlit    => outChannels(1),
      ready      => ready(1),
      full       => full(1),
      put        => put(1));

  Queue_two : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(2),
      flit_array => packet_i,
      outFlit    => outChannels(2),
      ready      => ready(2),
      full       => full(2),
      put        => put(2));

  Queue_three : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(3),
      flit_array => packet_i,
      outFlit    => outChannels(3),
      ready      => ready(3),
      full       => full(3),
      put        => put(3));

  Queue_four : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(4),
      flit_array => packet_i,
      outFlit    => outChannels(4),
      ready      => ready(4),
      full       => full(4),
      put        => put(4));

  Queue_five : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(5),
      flit_array => packet_i,
      outFlit    => outChannels(5),
      ready      => ready(5),
      full       => full(5),
      put        => put(5));

  Queue_six : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(6),
      flit_array => packet_i,
      outFlit    => outChannels(6),
      ready      => ready(6),
      full       => full(6),
      put        => put(6));

  Queue_seven : Queue
    port map (
      clk        => clk,
      reset      => reset,
      load       => load(7),
      flit_array => packet_i,
      outFlit    => outChannels(7),
      ready      => ready(7),
      full       => full(7),
      put        => put(7));

end behavioral;
