library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;
use ieee.numeric_std.all;
--use ieee.numeric_signed.all;
--use ieee.numeric_bit.all;

entity scheduler is

  port (
    clk           : in  std_logic;
    reset         : in  std_logic;
    inFlit        : in  std_logic_vector(16 downto 0);
    empty         : in  std_logic_vector(7 downto 0);
    get           : out std_logic_vector(7 downto 0);
    inChSel       : out std_logic_vector(2 downto 0);
    inChPri       : out priority;
    in_flit_array : out flitRegType);


end scheduler;

architecture behavioral of scheduler is

  ------------------------------------------------------------------------------
  -- These mask constants are used to differentiate the positions of the 
  ------------------------------------------------------------------------------
  constant GT_mask   : std_logic_vector(num_vc-1 downto 0) := "11000000";
  constant GP_mask   : std_logic_vector(num_vc-1 downto 0) := "00110000";
  constant BE_H_mask : std_logic_vector(num_vc-1 downto 0) := "00001100";
  constant BE_L_mask : std_logic_vector(num_vc-1 downto 0) := "00000011";


  signal get_check       : std_logic;
  signal received        : std_logic := '0';
  signal regs            : RegType;     -- priority table
  -- tag indicating which virtual channel has previously been selected for each
  -- priority type.
  type maskarray_type is array(3 downto 0) of std_logic_vector(num_vc-1 downto 0);
  signal mask_array      : maskarray_type;
  signal intermediateGet : std_logic_vector(num_vc-1 downto 0);
  signal intermediatePri : priority;
  signal intermediateSel : std_logic_vector(chSelect-1 downto 0);
  signal pk_received     : std_logic;

  type get_array is array(7 downto 0) of std_logic_vector(7 downto 0);



begin  -- behavioral
-------------------------------------------------------------------------------
-- Name: Priority Table for incoming virtual channels
-- Purpose: Incoming Virtual Channel Priority Assignment (Static)
-------------------------------------------------------------------------------
  regs(0) <= BE_L;
  regs(1) <= BE_L;
  regs(2) <= BE_H;
  regs(3) <= BE_H;
  regs(4) <= GP;
  regs(5) <= GP;
  regs(6) <= GT;
  regs(7) <= GT;

-------------------------------------------------------------------------------
-- Name: Virtual Channel Priority Scheduler
-- Purpose: Based on the priorities of the incoming virtual channels assigned
-- in the Priority Table and the empty bits for each channel, the scheduler
-- decides from which incoming virtual channel to select the next incoming packet.
-------------------------------------------------------------------------------
  mask_array(3) <= GT_mask;
  mask_array(2) <= GP_mask;
  mask_array(1) <= BE_H_mask;
  mask_array(0) <= BE_L_mask;



  Scheduler                   : process (mask_array, regs, empty)
    variable channel_selected : integer;
    variable pri              : integer range 0 to 3                := 0;
    variable ch_num           : integer range 0 to num_vc-1         := 0;
    variable contain_flit     : std_logic_vector(7 downto 0);
    variable temp_untagged    : std_logic_vector(num_vc-1 downto 0) := (others => '0');
    variable tag, tag_new     : std_logic_vector(num_vc-1 downto 0) := (others => '0');
    variable zero             : std_logic_vector(num_vc-1 downto 0) := (others => '0');
  begin  -- process
    -- The 'pri' for loop traverse through each priority level from highest
    -- priority to lowest priority.
    channel_selected                                                := 0;
    for pri in 3 downto 0 loop
      -- pri := 3;
      contain_flit                                                  := (others => '0');
      if channel_selected = 0 then
        for ch_num in 0 to num_vc-1 loop
          if regs(ch_num) = pri and empty(ch_num) = '0' then
            contain_flit(ch_num)                                    := '1';
            channel_selected                                        := channel_selected + 1;
          end if;
        end loop;
        next when channel_selected = 0;
        if temp_untagged /= zero then
          for ch_num in 0 to num_vc-1 loop
            if temp_untagged(ch_num) = '1' then
              intermediateGet         <= (others                               => '0');
              intermediateGet(ch_num) <= '1';
              temp_untagged(ch_num)                                 := '0';
              tag                                                   := tag and not mask_array(pri);
              tag(ch_num)                                           := '1';
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        elsif channel_selected = 1 then
          intermediateGet             <= contain_flit;
          tag                                                       := (tag and not mask_array(pri)) or contain_flit;
          for ch_num in 0 to num_vc-1 loop
            if contain_flit(ch_num) = '1' then
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        else
          temp_untagged                                             := contain_flit and not tag;
          for ch_num in 0 to num_vc-1 loop
            if temp_untagged(ch_num) = '1' then
              intermediateGet         <= (others                               => '0');
              intermediateGet(ch_num) <= '1';
              temp_untagged(ch_num)                                 := '0';
              tag                                                   := tag and not mask_array(pri);
              tag(ch_num)                                           := '1';
              intermediateSel         <= std_logic_vector(to_unsigned(ch_num, chSelect));
              intermediatePri         <= regs(ch_num);
              exit;
            end if;
          end loop;
        end if;
      end if;
    end loop;
    if channel_selected = 0 then
      intermediateGet                 <= (others                               => '0');
      intermediateSel <= (others => '0');
    end if;
  end process;

  process (clk, reset)
    variable insert_counter  : integer range 0 to num_vc                := 0;
    variable take_counter, i : integer range 0 to num_vc                := 0;
    variable count_flit      : integer range 0 to max_no_flit_in_packet := 0;  -- keeps track of the number of flits
    variable first           : boolean                                  := false;
    variable last            : boolean                                  := false;
    variable get_queue       : get_array;
  begin  -- process
    if reset = '1' then                 -- asynchronous reset (active high)
      insert_counter                                                    := 0;
      take_counter := 0;
      for i in 0 to 7 loop
        get_queue(i)                                                    := (others => '0');
      end loop;  -- i
      for i in 0 to max_no_flit_in_packet-1 loop
        in_flit_array(i) <= (others                                                => '0');
      end loop;  -- i 
      get                <= (others                                                => '0');
      inChSel            <= (others                                                => '0');
      inChPri            <= BE_L;
      received           <= '0';

    elsif clk'event and clk = '1' then  -- rising clock edge
      if count_flit = 0 then
        last := false;
      end if;
      if reset = '1' then
        count_flit   := 0;
        received                    <= '0';
      elsif get_check = '1' then
        count_flit   := count_flit + 1;
        if inFlit(flit_size-1) = '0' then
          if count_flit = 1 then
            first    := true;
          else
            first    := false;
          end if;
          received                  <= '0';
        elsif inFlit(flit_size-1) = '1' then
          last       := true;
          first      := false;
          received                  <= '1';
        end if;
      else
        received                    <= '0';
      end if;
      if count_flit = 0 then
        in_flit_array(count_flit+1) <= inFlit(flit_size-2 downto 0);
      else
        in_flit_array(count_flit-1) <= inFlit(flit_size-2 downto 0);
        if inFlit(flit_size-1) = '1' then
          count_flit := 0;
        end if;
      end if;
      if insert_counter = 0 then
        get_queue(insert_counter) := intermediateGet;
        insert_counter            := insert_counter + 1;
      elsif get_queue(insert_counter-1) /= intermediateGet then
        get_queue(insert_counter) := intermediateGet;
        insert_counter            := insert_counter + 1;
        if insert_counter = 8 then
          insert_counter          := 0;
        end if;
      end if;
      
      if first = true then
        get <= get_queue(take_counter);
        for i in 0 to num_vc -1 loop
          if get_queue(take_counter)(i) = '1' then
            inChSel <= std_logic_vector(to_unsigned(i, chSelect));
            inChPri <= regs(i);
          end if;
        end loop;  -- i
       --  get_queue(take_counter) := (others => '0');
        take_counter   := take_counter + 1;
        if take_counter = 8 then
          take_counter := 0;
        end if;
      end if;
    end if;
  end process;
  get_check <= intermediateGet(0) or intermediateGet(1) or intermediateGet(2) or intermediateGet(3) or intermediateGet(4) or intermediateGet(5)
               or intermediateGet(6) or intermediateGet(7);
  pk_received <= received;
end behavioral;



library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;

package scheduler_pkg is
  component scheduler
    port (
      clk           : in  std_logic;
      reset         : in  std_logic;
      inFlit        : in  std_logic_vector(16 downto 0);
      empty         : in  std_logic_vector(7 downto 0);
      get           : out std_logic_vector(7 downto 0);
      inChSel       : out std_logic_vector(2 downto 0);
      inChPri       : out Priority;
      in_flit_array : out flitRegType);
  end component;
end scheduler_pkg;

