library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;
-- contains 2 components
use work.regFile.all;
use work.compare.all;


entity vcScheduler is
  port (
    clk         : in  bit_t;
    reset       : in  bit_t;
    empty       : in  std_logic_vector(num_vc-1 downto 0);
    get         : out std_logic_vector(num_vc-1 downto 0);
    inChPri     : out priority;
    inChSel     : out std_logic_vector(chSelect-1 downto 0);
    wrChSel     : in  std_logic_vector(chSelect-1 downto 0);
    wrPri       : in  priority;
    wrEn        : in  bit_t;
    pk_received : in  bit_t);
end vcScheduler;

architecture behavioral of vcScheduler is
-- signal declarations
  signal vc0, vc1, vc2, vc3, vc4, vc5, vc6, vc7 : vcRecord;
  signal comp1, comp2, comp3, comp4             : vcRecord;
  signal comp2_1, comp2_2                       : vcRecord;
  signal comp3_1                                : vcRecord;

  signal tempGet : std_logic_vector(num_vc-1 downto 0);

  signal currentInChGet : std_logic_vector(num_vc-1 downto 0);
  signal currentInChSel : std_logic_vector(chSelect-1 downto 0);
  signal currentInChPri : priority;
  signal nextInChGet    : std_logic_vector(num_vc-1 downto 0);
  signal nextInChSel    : std_logic_vector(chSelect-1 downto 0);
  signal nextInChPri    : priority;

  signal RegWrite : bit_t;
  signal WrReg    : std_logic_vector(chSelect-1 downto 0);
  signal RdData   : RegType;
  signal WrData   : priority;

  signal count : integer range 0 to 1 := 0;


begin  -- behavioral

  priority_table : component regFile
    port map (
      clk      => clk,
      reset    => reset,
      RegWrite => RegWrite,
      WrReg    => WrReg,
      RdData   => RdData,
      WrData   => WrData);

  tableUpdate : process (wrEn, wrPri, wrChSel)
  begin  -- process tableUpdate
    if wrEn = '1' then
      WrData   <= wrPri;
      WrReg    <= wrChSel;
      RegWrite <= wrEn;
    else
      RegWrite <= '0';
    end if;
  end process tableUpdate;

  vc0.data  <= RdData(0);
  vc0.empty <= empty(0);
  vc0.pos   <= 0;
  vc1.data  <= RdData(1);
  vc1.empty <= empty(1);
  vc1.pos   <= 1;
  vc2.data  <= RdData(2);
  vc2.empty <= empty(2);
  vc2.pos   <= 2;
  vc3.data  <= RdData(3);
  vc3.empty <= empty(3);
  vc3.pos   <= 3;
  vc4.data  <= RdData(4);
  vc4.empty <= empty(4);
  vc4.pos   <= 4;
  vc5.data  <= RdData(5);
  vc5.empty <= empty(5);
  vc5.pos   <= 5;
  vc6.data  <= RdData(6);
  vc6.empty <= empty(6);
  vc6.pos   <= 6;
  vc7.data  <= RdData(7);
  vc7.empty <= empty(7);
  vc7.pos   <= 7;



  Comp_1and2 : component compare
    port map (
      input1 => vc0,
      input2 => vc1,
      output => comp1);

  Comp_3and4 : component compare
    port map (
      input1 => vc2,
      input2 => vc3,
      output => comp2);
  Comp_5and6 : component compare
    port map (
      input1 => vc4,
      input2 => vc5,
      output => comp3);
  Comp_7and8 : component compare
    port map (
      input1 => vc6,
      input2 => vc7,
      output => comp4);

  Comp2_1and2 : component compare
    port map (
      input1 => comp1,
      input2 => comp2,
      output => comp2_1);
  Comp2_3and4 : component compare
    port map (
      input1 => comp3,
      input2 => comp4,
      output => comp2_2);

  Comp3_1and2 : component compare
    port map (
      input1 => comp2_1,
      input2 => comp2_2,
      output => comp3_1);


  process (comp3_1, tempGet)
  begin  -- process
    for i in num_vc-1 downto 0 loop
      if i = comp3_1.pos and comp3_1.empty = '0' then
        tempGet(i) <= '1';
      else
        tempGet(i) <= '0';
      end if;
    end loop;
  end process;

  process (count, tempGet, comp3_1)
  begin  -- process
    if count = 0 then
      currentInChGet <= tempGet;
      currentInChSel <= std_logic_vector(to_unsigned(comp3_1.pos, chSelect));
      currentInChPri <= comp3_1.data;
      count          <= count + 1;
    else
      nextInChGet    <= tempGet;
      nextInChSel    <= std_logic_vector(to_unsigned(comp3_1.pos, chSelect));
      nextInChPri    <= comp3_1.data;
    end if;
  end process;

  process (pk_received, currentInChGet, currentInChSel, currentInChPri, nextInChGet, nextInChSel, nextInChPri)
  begin  -- process
    loop
      if pk_received /= '1' then
        get            <= currentInChGet;
        inChSel        <= currentInChSel;
        inChPri        <= currentInChPri;
      else
        get            <= nextInChGet;
        inChSel        <= nextInChSel;
        inChPri        <= nextInChPri;
        currentInChGet <= nextInChGet;
        currentInChSel <= nextInChSel;
        currentInChPri <= nextInChPri;
      end if;
    end loop;
  end process;

end behavioral;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.types.all;
-- contains 2 components
use work.regFile.all;
use work.compare.all;

package vcScheduler is
  component vcScheduler
    port (
      clk         : in  bit_t;
      reset       : in  bit_t;
      empty       : in  std_logic_vector(num_vc-1 downto 0);
      get         : out std_logic_vector(num_vc-1 downto 0);
      inChPri     : out priority;
      inChSel     : out std_logic_vector(chSelect-1 downto 0);
      wrChSel     : in  std_logic_vector(chSelect-1 downto 0);
      wrPri       : in  priority;
      wrEn        : in  bit_t;
      pk_received : in  bit_t);
  end component;
end vcScheduler;
